!function() {
  Function.prototype.method=function(name,func) { return this.prototype[name]=func, this }, 
  Function.method("inherits", function(superCtor) { return this.super_=superCtor, this.prototype=Object.create(superCtor.prototype, {constructor:{value:this,enumerable:!1,writable:!0,configurable:!0}}), this });
  ModbusClient=function(options, callback) {
    var ACTION={CONNECTING:0, CLOSED:1, IDLE:2, WAIT:3};
    var wsocket=null;
    var readRequestQueue=[];
    var writeRequestQueue=[];
    var curRequest={request:null, next:0};

    var settings = {
      automaticOpen: false,
      reconnectInterval: 1000,
      maxReconnectInterval: 30000,
      reconnectDecay: 1.5,
      timeoutInterval: 2000,
      maxReconnectAttempts: null,
      pollInterval: 500,
      receiveTimeout: 1000,
      modbusTcpPort: 502
    };

    if(!options) { options = {}; }
    for(var key in settings) {
      if(typeof options[key] != 'undefined') {
        this[key] = options[key];
      } else {
        this[key] = settings[key];
      }
    }

    this.url = location.hostname;
    this.port = this["modbusTcpPort"];

    this.callbackfun=null;
    if(callback && typeof callback === "function") this.callbackfun=callback;

    /* The number of attempted reconnects since starting, or the last successful connection. Read only. */
    this.reconnectAttempts = 0;

    /* The current state of the connection */
    this.readyConnectState = ACTION.CLOSED;
    this.failureResponse = 0;

    var self = this;
    var forcedClose = false;

    this.connect=function(reconnectAttempt) {
      if(reconnectAttempt) {
        if(this.maxReconnectAttempts && this.reconnectAttempts > this.maxReconnectAttempts) {
          return;
        }
      } else {
          this.reconnectAttempts = 0;
      }

      wsocket=new WebSocket("ws://" + this.url + ":" + this.port);
      wsocket.binaryType="arraybuffer";

      this.readyConnectState = ACTION.CONNECTING;

      var localWs = wsocket;
      var timeout = setTimeout(function() {
//console.debug('setTimeout', 'connection-timeout', localWs.url);
        localWs.close(1000);
      }, self.timeoutInterval);

      wsocket.onopen = function(event) {
//console.debug('onopen');

        clearTimeout(timeout);
        self.readyConnectState = ACTION.IDLE;
        self.reconnectAttempts = 0;
        //reconnectAttempt = false;

        self.failureResponse = 0

        curRequest.request=null;
        curRequest.next=0;

        setTimeout(function() {
          self.sendPacket();
        }, self.pollInterval); // pollInterval = 500ms
      };

      wsocket.onclose = function(event) {
//console.debug("onclose", event.code);

        clearTimeout(timeout);
        wsocket=null;

        if(curRequest.request && curRequest.request.getTimeout()) {
//console.debug("clearTimeout", curRequest.request.getTimeout());
          clearTimeout(curRequest.request.getTimeout());
          curRequest.request.setTimeout(null);
        }

        /* By default, forcedClose is false.
           If disconnect() is called -> forcedClose = true */
        if(forcedClose) {
          self.readyConnectState = ACTION.CLOSED;
        }
        else {
          self.readyConnectState = ACTION.CONNECTING;

          var timeout = self.reconnectInterval * Math.pow(self.reconnectDecay, self.reconnectAttempts);
//console.debug("reconnect timeout : ", timeout > self.maxReconnectInterval ? self.maxReconnectInterval : timeout);
          setTimeout(function() {
            self.reconnectAttempts++;
            self.connect(true);
          }, timeout > self.maxReconnectInterval ? self.maxReconnectInterval : timeout);
        }
      };

      wsocket.onmessage=function(event) {
        if(event.data.byteLength < 9) {
          console.debug("ModbusRequestManager","Wrong packet size.", event.data.byteLength);
        }

        if(event.data instanceof ArrayBuffer) {
          var mbap=new DataView(event.data, 0, 7), tid=mbap.getUint16(0);
          if(!curRequest.request) {
//console.error("ModbusRequestManager","No current request, strange!!");
            return;
          }
          if(curRequest.request.getId()!==tid) {
//console.error("ModbusRequestManager","CurrentRequest tid !== received tid",curRequest.request.getId(),tid);
            return;
          }

          var pdu=new DataView(event.data, 7, 2), fc=pdu.getUint8(0), exception=pdu.getUint8(1);
          if(fc & 0x80) {
            if(typeof self.callbackfun === "function") {
              self.callbackfun({id:tid,data:[]}, {errCode:"exception",code:exception});
            }
          }
          else {
            var byte_count=pdu.getUint8(1);
            var fc_data=[], count=curRequest.request.getCount();

            switch(fc) {
              case 0x01:
              case 0x02:
                var t,dv=new DataView(event.data, 9, byte_count);
                for(var i=0, j; i<byte_count; i++) {
                  for(t=dv.getUint8(i), j=0; j<8 && ((i<<3)+j<count); j++) {
                    fc_data.push(((t >> j) & 0x01));
                  }
                }
                break;

              case 0x03:
              case 0x04:
                for(var dv=new DataView(event.data, 9, byte_count), i=0; (byte_count>>1)>i; i++) {
                  fc_data.push(dv.getUint16(i<<1));
                }
                break;

              case 0x05:
              case 0x06:
                var _data=new DataView(event.data, 10, 2);
                fc_data.push(_data.getUint16(0));
                _data=null;
                break;
            }

            if(typeof self.callbackfun === "function") {
              if(!fc_data.length) self.callbackfun({id:tid,data:[]}, {errCode:"unknown_issue", code:0x01});
              else self.callbackfun({id:tid, data:fc_data});
            }
            
            self.failureResponse = 0;
          }

          pdu=null;
        }

        clearTimeout(curRequest.request.getTimeout());
        curRequest.request.setTimeout(null);

        setTimeout(function() {
          self.sendPacket();
        }, self.pollInterval);
      };

      wsocket.onerror=function(event) {
//console.debug("onerror", event.data);
        if(typeof self.callbackfun === "function") {
          self.callbackfun({id:{},data:[]}, {errCode:"faild"});
        }
      };
    };

    if(this.automaticOpen == true) {
      this.connect(false);
    }

    this.sendPacket=function() {
      if(!wsocket || wsocket.readyState !== WebSocket.OPEN) return;

      if(writeRequestQueue && writeRequestQueue.length) {
        curRequest.request=writeRequestQueue.shift();
      }
      else if(readRequestQueue && readRequestQueue.length) {
        curRequest.request=readRequestQueue[curRequest.next];
        if(++curRequest.next >= readRequestQueue.length) curRequest.next=0;
      }
      else curRequest.request=null;
//console.debug("sent", curRequest.request.getId());

      if(curRequest.request === null) { curRequest.next=0; return; }
      wsocket.send(curRequest.request.getPacket());

      var timeout_no=setTimeout(function() {
//console.debug("timeout", curRequest.request.getId(), typeof this.callbackfun === "function");
        if(++self.failureResponse >= 2) {
          if(typeof this.callbackfun === "function") {
            this.callbackfun({id:curRequest.request.getId(),data:[]}, {errCode:"timeout"});
          }
          clearTimeout(curRequest.request.getTimeout());
          curRequest.request.setTimeout(null);
          wsocket.close(1000);
        }
        else if(wsocket.readyState === WebSocket.OPEN) {
          wsocket.onmessage({data:[]});
        }
      }.bind(this), this.receiveTimeout);
      curRequest.request.setTimeout(timeout_no);
    };

    this.disconnect=function() {
//console.debug("disconnect", wsocket instanceof WebSocket);
      forcedClose = true;
      if(wsocket instanceof WebSocket) {
        wsocket.close(1000);
      }
    };

    this.getConnectState=function() {
      return wsocket && wsocket.readyState;
    };

    this.readCoils=function(start, count) {
      var tid=readRequestQueue.length;
      var request=new ModbusRequest(tid, 0x01, start, count);

      readRequestQueue.push(request);
      request.next=0;

      return tid;
    };

    this.readInputs=function(start, count) {
      var tid=readRequestQueue.length;
      var request=new ModbusRequest(tid, 0x02, start, count);

      readRequestQueue.push(request);
      request.next=0;

      return tid;
    };

    this.readHoldingRegisters=function(start, count) {
      var tid=readRequestQueue.length;
      var request=new ModbusRequest(tid, 0x03, start, count);

      readRequestQueue.push(request);
      request.next=0;

      return tid;
    };

    this.readInputRegisters=function(start, count) {
      var tid=readRequestQueue.length;
      var request=new ModbusRequest(tid, 0x04, start, count);

      readRequestQueue.push(request);
      request.next=0;

      return tid;
    };

    this.writeSingleCoil=function(address, value) {
      if(wsocket instanceof WebSocket === false) return false;

      var d = new Date(), tid=d.getMilliseconds();
      var request=new ModbusRequest(tid, 0x05, address, null, value);
      writeRequestQueue.push(request);

      if(curRequest.request === null) this.sendPacket();
      return tid;
    };

    this.writeSingleRegister=function(address, value) {
      if(wsocket instanceof WebSocket === false) return false;

      var d = new Date(), tid=d.getMilliseconds();
      var request=new ModbusRequest(tid, 0x06, address, null, value);
      writeRequestQueue.push(request);

      if(curRequest.request === null) this.sendPacket();
      return tid;
    };
  }

  function ModbusRequest(id, functionCode, start, count, value) {
    var packet=new ArrayBuffer(12), header=new DataView(packet, 0, 7), body=new DataView(packet, 7, 5), timeout=null;
    var init=function() {
      header.setUint16(0, id); // TID
      header.setUint16(2, 0); // PID
      header.setUint16(4, 6); // length
      header.setUint8(6, 127);
      body.setUint8(0, functionCode);
      body.setUint16(1, start);
      if(functionCode <= 0x04) body.setUint16(3, count);
      else if(functionCode == 0x05) body.setUint16(3, (!value?0:0xFF00));
      else if(functionCode == 0x06) body.setUint16(3, value);
    };
    this.getId=function() { return id; };
    this.getFunctionCode=function() { return functionCode; }
    this.getCount=function() { return count; }
    this.getPacket=function() { return packet; };
    this.getTimeout=function() { return timeout; };
    this.setTimeout=function(to) { return timeout=to, this };

    init();
  }
}();